/*
   (c) Copyright 2001-2011  The world wide DirectFB Open Source Community (directfb.org)
   (c) Copyright 2000-2004  Convergence (integrated media) GmbH

   All rights reserved.

   Written by Denis Oliver Kropp <dok@directfb.org>,
              Andreas Hundt <andi@fischlustig.de>,
              Sven Neumann <neo@directfb.org>,
              Ville Syrjälä <syrjala@sci.fi> and
              Claudio Ciccani <klan@users.sf.net>.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <config.h>

#include "CoreSurface.h"

extern "C" {
#include <directfb_util.h>

#include <direct/debug.h>
#include <direct/mem.h>
#include <direct/memcpy.h>
#include <direct/messages.h>

#include <fusion/conf.h>

#include <core/core.h>
}

D_DEBUG_DOMAIN( DirectFB_CoreSurface, "DirectFB/CoreSurface", "DirectFB CoreSurface" );

/*********************************************************************************************************************/

DFBResult
CoreSurface_SetConfig(
                    CoreSurface                               *obj,
                    const CoreSurfaceConfig                   *config
)
{
    if (!fusion_config->secure_fusion || dfb_core_is_master( core_dfb )) {
        DirectFB::ISurface_Real real( core_dfb, obj );

        return real.SetConfig( config );
    }

    DirectFB::ISurface_Requestor requestor( core_dfb, obj );

    return requestor.SetConfig( config );
}

DFBResult
CoreSurface_LockBuffer(
                    CoreSurface                               *obj,
                    CoreSurfaceBufferRole                      role,
                    CoreSurfaceAccessorID                      accessor,
                    CoreSurfaceAccessFlags                     access,
                    CoreSurfaceBufferLock                     *ret_lock
)
{
    if (!fusion_config->secure_fusion || dfb_core_is_master( core_dfb )) {
        DirectFB::ISurface_Real real( core_dfb, obj );

        return real.LockBuffer( role, accessor, access, ret_lock );
    }

    DirectFB::ISurface_Requestor requestor( core_dfb, obj );

    return requestor.LockBuffer( role, accessor, access, ret_lock );
}

DFBResult
CoreSurface_UnlockBuffer(
                    CoreSurface                               *obj,
                    CoreSurfaceBufferLock                     *lock
)
{
    if (!fusion_config->secure_fusion || dfb_core_is_master( core_dfb )) {
        DirectFB::ISurface_Real real( core_dfb, obj );

        return real.UnlockBuffer( lock );
    }

    DirectFB::ISurface_Requestor requestor( core_dfb, obj );

    return requestor.UnlockBuffer( lock );
}

DFBResult
CoreSurface_Flip(
                    CoreSurface                               *obj,
                    bool                                       swap
)
{
    if (!fusion_config->secure_fusion || dfb_core_is_master( core_dfb )) {
        DirectFB::ISurface_Real real( core_dfb, obj );

        return real.Flip( swap );
    }

    DirectFB::ISurface_Requestor requestor( core_dfb, obj );

    return requestor.Flip( swap );
}

DFBResult
CoreSurface_SetPalette(
                    CoreSurface                               *obj,
                    CorePalette                               *palette
)
{
    if (!fusion_config->secure_fusion || dfb_core_is_master( core_dfb )) {
        DirectFB::ISurface_Real real( core_dfb, obj );

        return real.SetPalette( palette );
    }

    DirectFB::ISurface_Requestor requestor( core_dfb, obj );

    return requestor.SetPalette( palette );
}

/*********************************************************************************************************************/

static FusionCallHandlerResult
CoreSurface_Dispatch( int           caller,   /* fusion id of the caller */
                     int           call_arg, /* optional call parameter */
                     void         *ptr, /* optional call parameter */
                     unsigned int  length,
                     void         *ctx,      /* optional handler context */
                     unsigned int  serial,
                     void         *ret_ptr,
                     unsigned int  ret_size,
                     unsigned int *ret_length )
{
    DirectFB::CoreSurfaceDispatch *dispatch = (DirectFB::CoreSurfaceDispatch*) ctx;

    dispatch->Dispatch( caller, call_arg, ptr, length, ret_ptr, ret_size, ret_length );

    return FCHR_RETURN;
}

void *CoreSurface_Init_Dispatch(
                    CoreDFB              *core,
                    CoreSurface          *obj,
                    FusionCall           *call
)
{
    DirectFB::CoreSurfaceDispatch *dispatch = new DirectFB::CoreSurfaceDispatch( core, new DirectFB::ISurface_Real(core, obj) );

    fusion_call_init3( call, CoreSurface_Dispatch, dispatch, core->world );

    return dispatch;
}

void  CoreSurface_Deinit_Dispatch(
                    void                 *_dispatch
)
{
    DirectFB::CoreSurfaceDispatch *dispatch = (DirectFB::CoreSurfaceDispatch*) _dispatch;

    delete dispatch;
}

/*********************************************************************************************************************/

namespace DirectFB {



DFBResult
ISurface_Requestor::SetConfig(
                    const CoreSurfaceConfig                   *config
)
{
    DFBResult           ret;
    CoreSurfaceSetConfig       *block = (CoreSurfaceSetConfig*) alloca( sizeof(CoreSurfaceSetConfig) );
    CoreSurfaceSetConfigReturn  return_block;

    D_DEBUG_AT( DirectFB_CoreSurface, "ISurface_Requestor::%s()\n", __FUNCTION__ );

    D_ASSERT( config != NULL );

    block->config = *config;

    ret = (DFBResult) CoreSurface_Call( obj, FCEF_NONE, CoreSurface_SetConfig, block, sizeof(CoreSurfaceSetConfig), &return_block, sizeof(return_block), NULL );
    if (ret) {
        D_DERROR( ret, "%s: CoreSurface_Call( CoreSurface_SetConfig ) failed!\n", __FUNCTION__ );
        return ret;
    }

    if (return_block.result) {
         D_DERROR( return_block.result, "%s: CoreSurface_SetConfig failed!\n", __FUNCTION__ );
         return return_block.result;
    }


    return DFB_OK;
}


DFBResult
ISurface_Requestor::LockBuffer(
                    CoreSurfaceBufferRole                      role,
                    CoreSurfaceAccessorID                      accessor,
                    CoreSurfaceAccessFlags                     access,
                    CoreSurfaceBufferLock                     *ret_lock
)
{
    DFBResult           ret;
    CoreSurfaceLockBuffer       *block = (CoreSurfaceLockBuffer*) alloca( sizeof(CoreSurfaceLockBuffer) );
    CoreSurfaceLockBufferReturn  return_block;

    D_DEBUG_AT( DirectFB_CoreSurface, "ISurface_Requestor::%s()\n", __FUNCTION__ );

    D_ASSERT( ret_lock != NULL );

    block->role = role;
    block->accessor = accessor;
    block->access = access;

    ret = (DFBResult) CoreSurface_Call( obj, FCEF_NONE, CoreSurface_LockBuffer, block, sizeof(CoreSurfaceLockBuffer), &return_block, sizeof(return_block), NULL );
    if (ret) {
        D_DERROR( ret, "%s: CoreSurface_Call( CoreSurface_LockBuffer ) failed!\n", __FUNCTION__ );
        return ret;
    }

    if (return_block.result) {
         D_DERROR( return_block.result, "%s: CoreSurface_LockBuffer failed!\n", __FUNCTION__ );
         return return_block.result;
    }

    *ret_lock = return_block.lock;

    return DFB_OK;
}


DFBResult
ISurface_Requestor::UnlockBuffer(
                    CoreSurfaceBufferLock                     *lock
)
{
    DFBResult           ret;
    CoreSurfaceUnlockBuffer       *block = (CoreSurfaceUnlockBuffer*) alloca( sizeof(CoreSurfaceUnlockBuffer) );
    CoreSurfaceUnlockBufferReturn  return_block;

    D_DEBUG_AT( DirectFB_CoreSurface, "ISurface_Requestor::%s()\n", __FUNCTION__ );

    D_ASSERT( lock != NULL );

    block->lock = *lock;

    ret = (DFBResult) CoreSurface_Call( obj, FCEF_NONE, CoreSurface_UnlockBuffer, block, sizeof(CoreSurfaceUnlockBuffer), &return_block, sizeof(return_block), NULL );
    if (ret) {
        D_DERROR( ret, "%s: CoreSurface_Call( CoreSurface_UnlockBuffer ) failed!\n", __FUNCTION__ );
        return ret;
    }

    if (return_block.result) {
         D_DERROR( return_block.result, "%s: CoreSurface_UnlockBuffer failed!\n", __FUNCTION__ );
         return return_block.result;
    }

    *lock = return_block.lock;

    return DFB_OK;
}


DFBResult
ISurface_Requestor::Flip(
                    bool                                       swap
)
{
    DFBResult           ret;
    CoreSurfaceFlip       *block = (CoreSurfaceFlip*) alloca( sizeof(CoreSurfaceFlip) );
    CoreSurfaceFlipReturn  return_block;

    D_DEBUG_AT( DirectFB_CoreSurface, "ISurface_Requestor::%s()\n", __FUNCTION__ );


    block->swap = swap;

    ret = (DFBResult) CoreSurface_Call( obj, FCEF_NONE, CoreSurface_Flip, block, sizeof(CoreSurfaceFlip), &return_block, sizeof(return_block), NULL );
    if (ret) {
        D_DERROR( ret, "%s: CoreSurface_Call( CoreSurface_Flip ) failed!\n", __FUNCTION__ );
        return ret;
    }

    if (return_block.result) {
         D_DERROR( return_block.result, "%s: CoreSurface_Flip failed!\n", __FUNCTION__ );
         return return_block.result;
    }


    return DFB_OK;
}


DFBResult
ISurface_Requestor::SetPalette(
                    CorePalette                               *palette
)
{
    DFBResult           ret;
    CoreSurfaceSetPalette       *block = (CoreSurfaceSetPalette*) alloca( sizeof(CoreSurfaceSetPalette) );
    CoreSurfaceSetPaletteReturn  return_block;

    D_DEBUG_AT( DirectFB_CoreSurface, "ISurface_Requestor::%s()\n", __FUNCTION__ );

    D_ASSERT( palette != NULL );

    block->palette_id = palette->object.id;

    ret = (DFBResult) CoreSurface_Call( obj, FCEF_NONE, CoreSurface_SetPalette, block, sizeof(CoreSurfaceSetPalette), &return_block, sizeof(return_block), NULL );
    if (ret) {
        D_DERROR( ret, "%s: CoreSurface_Call( CoreSurface_SetPalette ) failed!\n", __FUNCTION__ );
        return ret;
    }

    if (return_block.result) {
         D_DERROR( return_block.result, "%s: CoreSurface_SetPalette failed!\n", __FUNCTION__ );
         return return_block.result;
    }


    return DFB_OK;
}

/*********************************************************************************************************************/

DFBResult
CoreSurfaceDispatch::Dispatch( FusionID      caller,
                                int           method,
                                void         *ptr,
                                unsigned int  length,
                                void         *ret_ptr,
                                unsigned int  ret_size,
                                unsigned int *ret_length )
{
    D_UNUSED
    DFBResult ret;

    D_DEBUG_AT( DirectFB_CoreSurface, "CoreSurfaceDispatch::%s()\n", __FUNCTION__ );

    switch (method) {
        case CoreSurface_SetConfig: {
            D_UNUSED
            CoreSurfaceSetConfig       *args        = (CoreSurfaceSetConfig *) ptr;
            CoreSurfaceSetConfigReturn *return_args = (CoreSurfaceSetConfigReturn *) ret_ptr;

            D_DEBUG_AT( DirectFB_CoreSurface, "=-> CoreSurface_SetConfig\n" );

            return_args->result = real->SetConfig( &args->config );
            if (return_args->result == DFB_OK) {
            }

            *ret_length = sizeof(CoreSurfaceSetConfigReturn);

            return DFB_OK;
        }

        case CoreSurface_LockBuffer: {
            D_UNUSED
            CoreSurfaceLockBuffer       *args        = (CoreSurfaceLockBuffer *) ptr;
            CoreSurfaceLockBufferReturn *return_args = (CoreSurfaceLockBufferReturn *) ret_ptr;

            D_DEBUG_AT( DirectFB_CoreSurface, "=-> CoreSurface_LockBuffer\n" );

            return_args->result = real->LockBuffer( args->role, args->accessor, args->access, &return_args->lock );
            if (return_args->result == DFB_OK) {
            }

            *ret_length = sizeof(CoreSurfaceLockBufferReturn);

            return DFB_OK;
        }

        case CoreSurface_UnlockBuffer: {
            D_UNUSED
            CoreSurfaceUnlockBuffer       *args        = (CoreSurfaceUnlockBuffer *) ptr;
            CoreSurfaceUnlockBufferReturn *return_args = (CoreSurfaceUnlockBufferReturn *) ret_ptr;

            D_DEBUG_AT( DirectFB_CoreSurface, "=-> CoreSurface_UnlockBuffer\n" );

            return_args->result = real->UnlockBuffer( &args->lock );
            if (return_args->result == DFB_OK) {
                return_args->lock = args->lock;
            }

            *ret_length = sizeof(CoreSurfaceUnlockBufferReturn);

            return DFB_OK;
        }

        case CoreSurface_Flip: {
            D_UNUSED
            CoreSurfaceFlip       *args        = (CoreSurfaceFlip *) ptr;
            CoreSurfaceFlipReturn *return_args = (CoreSurfaceFlipReturn *) ret_ptr;

            D_DEBUG_AT( DirectFB_CoreSurface, "=-> CoreSurface_Flip\n" );

            return_args->result = real->Flip( args->swap );
            if (return_args->result == DFB_OK) {
            }

            *ret_length = sizeof(CoreSurfaceFlipReturn);

            return DFB_OK;
        }

        case CoreSurface_SetPalette: {
    CorePalette *palette = NULL;
            D_UNUSED
            CoreSurfaceSetPalette       *args        = (CoreSurfaceSetPalette *) ptr;
            CoreSurfaceSetPaletteReturn *return_args = (CoreSurfaceSetPaletteReturn *) ret_ptr;

            D_DEBUG_AT( DirectFB_CoreSurface, "=-> CoreSurface_SetPalette\n" );

            ret = (DFBResult) CorePalette_Lookup( core, args->palette_id, &palette );
            if (ret) {
                 D_DERROR( ret, "%s: Looking up palette by ID %u failed!\n", __FUNCTION__, args->palette_id );
                 return_args->result = ret;
                 return DFB_OK;
            }

            return_args->result = real->SetPalette( palette );
            if (return_args->result == DFB_OK) {
            }

            *ret_length = sizeof(CoreSurfaceSetPaletteReturn);

            if (palette)
                CorePalette_Unref( palette );

            return DFB_OK;
        }

    }

    return DFB_NOSUCHMETHOD;
}

}
